'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});
exports.dndWrapSource = dndWrapSource;
exports.dndWrapTarget = dndWrapTarget;
exports.dndWrapRoot = dndWrapRoot;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _reactDnd = require('react-dnd');

var _reactDndHtml5Backend = require('react-dnd-html5-backend');

var _reactDndHtml5Backend2 = _interopRequireDefault(_reactDndHtml5Backend);

var _treeDataUtils = require('./tree-data-utils');

var _memoizedTreeDataUtils = require('./memoized-tree-data-utils');

var nodeDragSource = {
  beginDrag: function beginDrag(props) {
    props.startDrag(props);

    return {
      node: props.node,
      parentNode: props.parentNode,
      path: props.path,
      treeIndex: props.treeIndex
    };
  },

  endDrag: function endDrag(props, monitor) {
    props.endDrag(monitor.getDropResult());
  },

  isDragging: function isDragging(props, monitor) {
    var dropTargetNode = monitor.getItem().node;
    var draggedNode = props.node;

    return draggedNode === dropTargetNode;
  }
};

function getTargetDepth(dropTargetProps, monitor) {
  var dropTargetDepth = 0;
  var draggedItem = monitor.getItem();
  var rowAbove = dropTargetProps.getPrevRow();
  if (rowAbove) {
    // Limit the length of the path to the deepest possible
    dropTargetDepth = Math.min(rowAbove.path.length, dropTargetProps.path.length);
  }

  var blocksOffset = Math.round(monitor.getDifferenceFromInitialOffset().x / dropTargetProps.scaffoldBlockPxWidth);

  var targetDepth = Math.min(dropTargetDepth, Math.max(0, draggedItem.path.length + blocksOffset - 1));

  // If a maxDepth is defined, constrain the target depth
  if (typeof dropTargetProps.maxDepth !== 'undefined' && dropTargetProps.maxDepth !== null) {
    var draggedNode = monitor.getItem().node;
    var draggedChildDepth = (0, _treeDataUtils.getDepth)(draggedNode);

    targetDepth = Math.min(targetDepth, dropTargetProps.maxDepth - draggedChildDepth - 1);
  }

  return targetDepth;
}

function canDrop(dropTargetProps, monitor) {
  if (!monitor.isOver()) {
    return false;
  }

  var rowAbove = dropTargetProps.getPrevRow();
  var abovePath = rowAbove ? rowAbove.path : [];
  var aboveNode = rowAbove ? rowAbove.node : {};
  var targetDepth = getTargetDepth(dropTargetProps, monitor);

  // Cannot drop if we're adding to the children of the row above and
  //  the row above is a function
  if (targetDepth >= abovePath.length && typeof aboveNode.children === 'function') {
    return false;
  }

  if (typeof dropTargetProps.customCanDrop === 'function') {
    var node = monitor.getItem().node;
    var addedResult = (0, _memoizedTreeDataUtils.memoizedInsertNode)({
      treeData: dropTargetProps.treeData,
      newNode: node,
      depth: targetDepth,
      getNodeKey: dropTargetProps.getNodeKey,
      minimumTreeIndex: dropTargetProps.listIndex,
      expandParent: true
    });

    return dropTargetProps.customCanDrop({
      node: node,
      prevPath: monitor.getItem().path,
      prevParent: monitor.getItem().parentNode,
      prevTreeIndex: monitor.getItem().treeIndex,
      nextPath: addedResult.path,
      nextParent: addedResult.parentNode,
      nextTreeIndex: addedResult.treeIndex
    });
  }

  return true;
}

var nodeDropTarget = {
  drop: function drop(dropTargetProps, monitor) {
    return {
      node: monitor.getItem().node,
      path: monitor.getItem().path,
      minimumTreeIndex: dropTargetProps.treeIndex,
      depth: getTargetDepth(dropTargetProps, monitor)
    };
  },

  hover: function hover(dropTargetProps, monitor) {
    var targetDepth = getTargetDepth(dropTargetProps, monitor);
    var draggedNode = monitor.getItem().node;
    var needsRedraw =
    // Redraw if hovered above different nodes
    dropTargetProps.node !== draggedNode ||
    // Or hovered above the same node but at a different depth
    targetDepth !== dropTargetProps.path.length - 1;

    if (!needsRedraw) {
      return;
    }

    dropTargetProps.dragHover({
      node: draggedNode,
      path: monitor.getItem().path,
      minimumTreeIndex: dropTargetProps.listIndex,
      depth: targetDepth
    });
  },

  canDrop: canDrop
};

function nodeDragSourcePropInjection(connect, monitor) {
  return {
    connectDragSource: connect.dragSource(),
    connectDragPreview: connect.dragPreview(),
    isDragging: monitor.isDragging(),
    didDrop: monitor.didDrop()
  };
}

function nodeDropTargetPropInjection(connect, monitor) {
  var dragged = monitor.getItem();
  return {
    connectDropTarget: connect.dropTarget(),
    isOver: monitor.isOver(),
    canDrop: monitor.canDrop(),
    draggedNode: dragged ? dragged.node : null
  };
}

function dndWrapSource(el, type) {
  return (0, _reactDnd.DragSource)(type, nodeDragSource, nodeDragSourcePropInjection)(el);
}

function dndWrapTarget(el, type) {
  return (0, _reactDnd.DropTarget)(type, nodeDropTarget, nodeDropTargetPropInjection)(el);
}

function dndWrapRoot(el) {
  return (0, _reactDnd.DragDropContext)(_reactDndHtml5Backend2['default'])(el);
}